/// <reference types="svelte" />
import { Map as ReactiveMap } from 'svelte/reactivity';
import { type TurboQuery, type TurboQueryOptions } from 'turbo-query';
export declare const setQueryContext: () => {
    cacheFactory: ReactiveMap<unknown, unknown>;
    depsFactory: {};
    originalResponseStore: Map<any, any>;
    endHook: () => void;
    instanceQueriesMetaStore: Map<any, any>;
};
export declare const getQueryContext: <TMerged, Deps, TData, TDeps>() => {
    cacheFactory: Map<string, {
        expiresAt: Date;
        item: TMerged;
    }>;
    depsFactory: {
        [x: string]: Set<string>;
    };
    instanceQueriesMetaStore: Map<string, {
        count: number[];
        names: Set<string>;
    }>;
    endHook: (props: {
        error?: Error;
        data?: TData;
    }) => void;
    originalResponseStore: Map<string, Map<TDeps, TData>>;
};
export declare const setQueryInstanceContext: () => Map<any, any>;
export declare const getQueryInstanceContext: () => Map<string, TurboQuery>;
export declare const setQueryCacheContext: () => Map<any, any>;
export declare const getQueryCacheContext: () => Map<string, {
    expiresAt: Date;
    item: unknown;
}>;
export declare const setQueryDepsStoreContext: () => {};
export declare const getQueryDepsStoreContext: () => {
    [x: string]: Set<unknown>;
};
type id = string | number | object;
export type normalized<key, value> = {
    ids: key[];
    /**
     * @description  typically the value contains an object keyed by id
     * @example {blog_id_1: blog_value,blog_id_2:blog_value,...}
     */
    value: object | Map<id, value>;
};
export declare function createQuery<TData, TError, TDeps, TMerged, TFinal>(props: {
    fetcher: (args: TDeps, abort: AbortSignal) => Promise<TData>;
    cacheKey: string;
    name?: string;
    deps: () => TDeps;
    /**
     *
     * @param data original response from provided fetcher
     * @description  get unique id from the response, which can be used later to pick item from cache
     * @returns
     */
    getUniqueIds?: (data: TData) => (string | number)[];
    normalizer?: () => {};
    /**
     *
     * @description returned merged result will replace current cached item
     * @param
     * @returns
     */
    mergeFn: (args: {
        Cached?: {
            expiresAt: Date;
            item?: TMerged;
        };
        /**
         * @description sometime the original response is expired, so we need to fetch new response
         */
        originalResponse: TData;
        /**
         * @description sometime the original response is expired, so we need to fetch new response, prevResponse is the expired response
         */
        prevResponse?: TData;
    }) => TMerged;
    /**
     * @description sometime the data you want is already in the cache, try search cache(mergeFn will be called on the fake response) then update response with actual fetch
     * @returns {TData}
     */
    optimisticReturn?: (args: {
        Cached: {
            expiresAt: Date;
            item: TMerged;
        };
    }) => TData;
    pickFn: (args: {
        Cached: {
            expiresAt: Date;
            item: TMerged;
        };
        originalResponse: TData;
        ids: (string | number)[];
    }) => TFinal;
    onSuccess?: (data: {
        picked: TFinal;
        originalResponse: TData;
        cache?: {
            expiresAt: Date;
            item: TMerged;
        };
    }) => void;
    onError?: (err: TError) => void;
    queryOptions?: TurboQueryOptions;
    initialData?: TData;
}): () => {
    data: TFinal | undefined;
    isError: boolean;
    isRefetching: boolean;
    isLoading: boolean;
    isAborted: boolean;
    isSuccess: boolean;
    instance: TurboQuery;
    cacheInstance: () => {
        expiresAt: Date;
        item: TMerged;
    } | undefined;
    refetch: () => Promise<TData | undefined>;
    error?: Error | undefined;
};
export {};
