import { getContext, onMount, setContext, tick, untrack } from 'svelte';
import { Map as ReactiveMap } from 'svelte/reactivity';
import { createTurboQuery } from 'turbo-query';
import stringify from 'fast-json-stable-stringify';
import { merge, pickBy, prop } from 'remeda';
function log() { }
/**
 * @description store instance query for query deduplication
 */
const instanceFactory = new Map();
/**
 * @description store deps change over time , for query deduplication based on cache key on a app level for all queries with same key
 */
const DepsFactory = {};
/**
 * @description store global cache for a cache key, reactive map so that all UI can be updated
 */
const globalCacheFactory = new Map();
const QUERY_CLIENT_TURBO = 'turbo query context';
const QUERY_CACHE_CLIENT_TURBO = 'turbo query cache context';
const QUERY_CONTEXT_INSTANCE = 'turbo query instance context';
const QUERY_DEPS_STORE_CLIENT_TURBO = 'turbo query deps store context';
export const setQueryContext = () => {
    return setContext(QUERY_CLIENT_TURBO, {
        cacheFactory: new ReactiveMap(),
        depsFactory: {},
        //instanceFactory: new Map(),
        //instanceQueries: new Map(),
        originalResponseStore: new Map(),
        endHook: () => console.log('query ended'),
        instanceQueriesMetaStore: new Map()
    });
};
export const getQueryContext = () => {
    return getContext(QUERY_CLIENT_TURBO);
};
export const setQueryInstanceContext = () => {
    return setContext(QUERY_CONTEXT_INSTANCE, new Map());
};
export const getQueryInstanceContext = () => {
    return getContext(QUERY_CONTEXT_INSTANCE);
};
export const setQueryCacheContext = () => {
    return setContext(QUERY_CACHE_CLIENT_TURBO, new Map());
};
export const getQueryCacheContext = () => {
    return getContext(QUERY_CACHE_CLIENT_TURBO);
};
export const setQueryDepsStoreContext = () => {
    return setContext(QUERY_DEPS_STORE_CLIENT_TURBO, {});
};
export const getQueryDepsStoreContext = () => {
    return getContext(QUERY_DEPS_STORE_CLIENT_TURBO);
};
export function createQuery(props) {
    /* 	const cacheFactory = getQueryCacheContext();
    const depsFactory = getQueryDepsStoreContext();
    const instanceFactory = getQueryInstanceContext(); */
    const { cacheFactory, depsFactory, instanceQueriesMetaStore, endHook, originalResponseStore } = getQueryContext();
    if (!depsFactory[props.cacheKey]) {
        depsFactory[props.cacheKey] = new Set();
    }
    let depsHistory = depsFactory[props.cacheKey];
    let ids = [];
    let hash = $derived(stringify(props.deps()));
    const queryInstance = createTurboQuery({
        itemsCache: {
            get(key) {
                const old_cache = cacheFactory.get(key);
                console.info('Get, check if item exists in cache', key);
                if (!old_cache) {
                    return;
                }
                if (old_cache.expiresAt.getTime() <= Date.now()) {
                    console.log('Get Item from cached But expired');
                }
                const original = originalResponse.get(hash);
                console.log('GET originalResponse', Array.from(originalResponse), originalResponse.get(hash));
                if (!original) {
                    // missing original response, need to refetch
                    console.log('missing original response, refetching...');
                    return;
                }
                result.data = props.pickFn({
                    Cached: old_cache,
                    originalResponse: original,
                    ids: props?.getUniqueIds?.(original) ?? []
                });
                console.log('GET originalResponse', JSON.stringify(result.data));
                clearTimer();
                result.isLoading = false;
                result.isSuccess = true;
                return old_cache;
            },
            set(key, value) {
                const cache = cacheFactory.get(key);
                const copied = JSON.parse(JSON.stringify(value.item));
                const prev_original = originalResponse.get(stringify(props.deps()));
                console.log('Setting cache', originalResponse, stringify(props.deps()));
                originalResponse.set(stringify(props.deps()), copied); // create a copy so that when cache is changed, originalResponse wont change
                console.time('Set cache');
                const temp = props.mergeFn({
                    Cached: cache,
                    originalResponse: copied,
                    prevResponse: prev_original
                });
                if (!cache?.item) {
                    let reactive_output = $state({ expiresAt: value.expiresAt, item: temp });
                    cacheFactory.set(key, reactive_output);
                }
                else {
                    merge(cache.item, temp);
                    cacheFactory.set(key, cache);
                }
                console.timeEnd('Set cache');
            },
            keys() {
                return cacheFactory.keys();
            },
            delete(key) {
                //console.log('delete ', key);
                return cacheFactory.delete(key);
            }
        }
    });
    let temp_orig = originalResponseStore.get(props.cacheKey);
    let originalResponse = temp_orig ?? new Map();
    if (!temp_orig)
        originalResponseStore.set(props.cacheKey, originalResponse);
    // increment count
    const meta = instanceQueriesMetaStore.get(props.cacheKey) ?? { count: [], names: new Set() };
    meta.count.push(1); // avoid race condition with push instead of inc
    instanceQueriesMetaStore.set(props.cacheKey, meta);
    function getResponseHistory() {
        return originalResponse.get(stringify(props.deps()));
    }
    if (props.initialData) {
        queryInstance.caches().items.set(props.cacheKey, {
            expiresAt: new Date(Date.now() + (props?.queryOptions?.expiration?.(props.initialData) ?? 0)),
            item: props.initialData
        });
        let k = stringify(props.deps());
        //	originalResponse.set(k, props.initialData);
        //depsHistory.add(k);
    }
    let result = $state({
        isAborted: false,
        isRefetching: false,
        data: undefined,
        queryInstance: queryInstance,
        isLoading: true,
        isSuccess: false,
        isError: false,
        instance: queryInstance,
        refetch: () => {
            return get(true);
        },
        cacheInstance: () => cacheFactory.get(props.cacheKey)
    });
    const optimisticReturn = props?.optimisticReturn?.({ Cached: cacheFactory.get(props.cacheKey) });
    if (optimisticReturn) {
        result.data = props.pickFn({
            Cached: cacheFactory.get(props.cacheKey),
            originalResponse: optimisticReturn
        });
    }
    async function get(force = false) {
        const result = await queryInstance?.query(props.cacheKey, {
            //@ts-expect-error idk
            fetcher: async (key, { signal }) => {
                const res = await props.fetcher(props.deps(), signal);
                clearTimer();
                return res;
            },
            ...(props.queryOptions ?? {}),
            fresh: force
        });
        return result;
    }
    let timer = { id: null };
    function clearTimer() {
        //console.log('clear time');
        if (timer.id)
            clearTimeout(timer.id);
    }
    $effect(() => {
        console.log('rerun eff', props.deps());
        console.log(props.name);
        if (props.deps() && hash) {
            untrack(() => {
                clearTimer();
                const has = depsHistory.has(hash);
                console.log('deps updated', has);
                if (!has) {
                    console.log('a new fetch', hash, Array.from(depsHistory), has);
                    //  a new query
                    depsHistory.add(hash);
                }
                result.isLoading = true;
                result.isSuccess = false;
                get(!has);
            });
        }
    });
    $effect.root(() => {
        const unsub = [];
        unsub.push(queryInstance.subscribe(props.cacheKey, 'resolved', function (payload) {
            result['isError'] = false;
            console.log('resolved query', payload.detail, originalResponse.keys());
            clearTimer();
            const res = getResponseHistory();
            const val = props.pickFn({
                Cached: cacheFactory.get(props.cacheKey),
                originalResponse: res,
                ids: ids
            });
            result.data = val;
            result.isRefetching = false;
            result.error = undefined;
            result.isAborted = false;
            result.isLoading = false;
            result.isSuccess = true;
            console.log('resolved  merged data ', JSON.stringify(val, null));
            props.onSuccess?.({
                picked: val,
                originalResponse: res,
                cache: cacheFactory.get(props.cacheKey)
            });
            endHook?.({ data: payload.detail });
        }));
        unsub.push(queryInstance.subscribe(props.cacheKey, 'aborted', function (payload) {
            result.isAborted = true;
            //	result.data = payload;
        }));
        unsub.push(queryInstance.subscribe(props.cacheKey, 'refetching', function (payload) {
            result.isRefetching = true; //expired or deps changed
            console.log('refetching', payload);
        }));
        unsub.push(queryInstance.subscribe(props.cacheKey, 'error', function (payload) {
            result.isError = true;
            result.error = payload.detail;
            result.isSuccess = false;
            props.onError?.(payload.detail);
            endHook({ error: payload.detail });
        }));
        unsub.push(queryInstance.subscribe(props.cacheKey, 'forgotten', function (payload) {
            //expired
            console.log('cache invalidated', payload);
            get(true);
        }));
        return () => {
            console.log('cleanup');
            queryInstance.abort();
            unsub.forEach((v) => v());
            const meta = instanceQueriesMetaStore.get(props.cacheKey) ?? { count: [] };
            meta.count.push(-1);
            instanceQueriesMetaStore.set(props.cacheKey, meta);
        };
    });
    const res = $derived({
        ...result
        //data: cacheFactory.get(props.cacheKey)?.item as TFinal
    });
    /* $effect(() => {
        console.log('res changed', res, Object.keys(cacheFactory.get(props.cacheKey) ?? {}));
    }); */
    //$inspect('inspecting', res).with(console.log);
    return () => result;
}
